import time
import json


class OpenConversationId:
    """OpenConversationId类，用于标识对象"""
    
    openConversationId: str
    """OpenConversationId值"""
    
    name: str
    """用于标识群名称，可能为无"""
    
    group_id: int
    """用于标识群ID(框架内显示ID)，可能为无"""
    
    def __init__(self, openConversationId, name="未知会话", group_id=0):
        self.openConversationId = openConversationId
        self.name = name
        self.group_id = group_id
    
    def __str__(self):
        return self.openConversationId
    
    def __int__(self):
        return self.group_id


class AccessToken:
    token: str
    
    expired: int = 0
    
    def __init__(self, accessToken: str = None, expireTime: int = 0):
        self.token = accessToken
        self.expired = expireTime if expireTime > 1600000000 else int(time.time()) + expireTime
    
    def __str__(self):
        return self.token
    
    def __int__(self):
        return self.expired
    
    @property
    def ok(self):
        return time.time() < self.expired


class TimeStamp:
    timestamp: int
    
    def __init__(self, timeStamp):
        self.timestamp = timeStamp
        
        
class Response:
    
    ok: bool = None
    
    url: str
    
    text: str
    
    sendData: dict
    
    recallType: str
    
    recallOpenConversationId: str = None
    
    @property
    def json(self):
        return json.loads(self.text)
    
    def __bool__(self):
        if self.ok:
            return True
        return False



import copy
import hashlib
import json
import os
from pathlib import Path
from typing import Union, BinaryIO

from ..model import Member


class File:
    type: "File"
    
    file: BinaryIO
    
    size: int
    
    fileType: str
    
    mediaId = None
    
    def __init__(self, file: Union[Path, BinaryIO, str] = None):
        if file:
            if not isinstance(file, BinaryIO):
                f = open(file, 'rb')
                self.file = f
                f.seek(0, os.SEEK_END)
                self.size = f.tell()
                f.seek(0)
            else:
                self.file = file
                self.size = len(self.file.read())
        self.mediaId = None
        self.fileName = None
        self.fileType = 'file'
    
    @property
    def template(self):
        return {
            'msgKey'  : "sampleImageMsg",
            'msgParam': json.dumps({
                'mediaId' : self.mediaId,
                'fileName': self.fileName,
                'fileType': self.fileType
            })
        }
    
    @property
    def data(self):
        raise TypeError("Image cannot be used in Webhook!")


class BaseElement:
    type: "BaseElement"
    
    data: dict
    
    template: dict


class Link(BaseElement):
    
    def __init__(self, url: str, title: str = "[Link]", text: str = "", pic_url: str = ""):
        """链接
        
        Args:
            title: 链接标题
            text: 链接简介
            url: 链接
            pic_url: 图片链接
        """
        self.url = url
        self.title = title
        self.text = text
        self.pic_url = pic_url if pic_url else url
        self.data = {
            "msgtype": "link",
            "link"   : {
                "title"     : str(self.title),
                "text"      : str(self.text),
                "picUrl"    : _link_detect(str(self.pic_url)),
                "messageUrl": _link_detect(str(self.url))
            }
        }
        self.template = {
            "msgKey"  : "sampleLink",
            "msgParam": json.dumps(self.data["link"])
        }
    
    def __str__(self):
        return f"[Link]({self.url})"


class Image(File):
    
    def __init__(self, file: Union[Path, BinaryIO, str] = None):
        super().__init__(file)
        if file:
            if not isinstance(file, BinaryIO):
                f = open(file, 'rb')
                self.file = f
                f.seek(0, os.SEEK_END)
                self.size = f.tell()
                f.seek(0)
            else:
                self.file = file
                self.size = len(self.file.read())
        self.fileType = 'image'
        self.mediaId = None
    
    @property
    def template(self):
        return {
            'msgKey'  : "sampleImageMsg",
            'msgParam': json.dumps({
                'photoURL': self.mediaId
            })
        }
    
    @property
    def data(self):
        raise TypeError("Image cannot be used in Webhook!")


class Audio(File):
    
    def __init__(self, file: Union[Path, BinaryIO, str] = None):
        super().__init__(file)
        if file:
            if not isinstance(file, BinaryIO):
                f = open(file, 'rb')
                self.file = f
                f.seek(0, os.SEEK_END)
                self.size = f.tell()
                f.seek(0)
            else:
                self.file = file
                self.size = len(self.file.read())
        self.fileType = 'voice'
        self.mediaId = None
        self.duration = None
    
    @property
    def template(self):
        return {
            'msgKey'  : "sampleAudio",
            'msgParam': json.dumps({
                'mediaId' : self.mediaId,
                'duration': self.duration
            })
        }
    
    @property
    def data(self):
        raise TypeError("Audio cannot be used in Webhook!")


class Video(File):
    
    def __init__(self, file: Union[Path, BinaryIO, str] = None):
        super().__init__(file)
        if file:
            if not isinstance(file, BinaryIO):
                f = open(file, 'rb')
                self.file = f
                f.seek(0, os.SEEK_END)
                self.size = f.tell()
                f.seek(0)
            else:
                self.file = file
                self.size = len(self.file.read())
        self.fileType = 'video'
        self.mediaId = None
        self.videoMediaId = None
        self.videoType = 'mp4'
        self.duration = None
        self.picMediaId = None
    
    @property
    def template(self):
        return {
            'msgKey'  : "sampleVideo",
            'msgParam': json.dumps({
                'duration'    : self.duration,
                'videoMediaId': self.mediaId,
                'videoType'   : self.videoType,
                'picMediaId'  : self.picMediaId
            })
        }
    
    @property
    def data(self):
        raise TypeError("Video cannot be used in Webhook!")


class Markdown(BaseElement):
    
    def __init__(self, text: str, title: str = "[Markdown]"):
        """发送Markdown形式的消息
        
        Args:
            title: 标题(似乎没啥用)
            text: Markdown文本
        """
        self.text = text
        self.title = title
        self.data = {
            "msgtype" : "markdown",
            "markdown": {
                "title": str(self.title),
                "text" : str(self.text),
            }
        }
        self.template = {
            "msgKey"  : "sampleMarkdown",
            "msgParam": json.dumps(self.data["markdown"])
        }
    
    def __str__(self):
        return "[Markdown]"


class ActionCard(BaseElement):
    
    def __init__(self, text: str, button: list, title: str = "[ActionCard]", orientation: int = 0):
        """发送ActionCard消息

        Args:
            title: 标题(似乎没啥用)
            text: Markdown文本
            button: 按钮列表，即使只有一个也要使用[[title, url], ...]的格式
            orientation: 排列方向，0为竖向，1为横向，建议横向不超过5个字
        """
        self.orientation = orientation
        self.button = button
        self.text = text
        self.title = title
        self.data = {
            "msgtype"   : "actionCard",
            "actionCard": {
                "title"         : str(self.title),
                "text"          : str(self.text),
                "btnOrientation": str(self.orientation)
            }
        }
        if len(self.button) == 2 and self.orientation:
            raise TypeError(
                "Only 2 bottoms are allowed when orientation is 1, but %s bottoms given!" % len(self.button))
        if len(self.button) > 5:
            raise TypeError("Only 5 bottoms are allowed but %s bottoms given!" % len(self.button))
        if len(self.button) == 1:
            self.data['actionCard']['singleTitle'] = str(self.button[0][0])
            self.data['actionCard']['singleURL'] = _link_detect(str(self.button[0][1]))
            self.data['actionCard'].pop('btnOrientation')
            self.template = {
                "msgKey"  : "sampleActionCard",
                "msgParam": json.dumps(self.data["actionCard"])
            }
        else:
            self.data['actionCard']['btns'] = []
            for b in self.button:
                if b[0] and b[1]:
                    self.data['actionCard']['btns'].append(
                        {"title": str(b[0]), "actionURL": _link_detect(str(b[1]))})
            if not self.orientation:
                if len(self.button) == 2:
                    self.template = {
                        "msgKey"  : "sampleActionCard2",
                        "msgParam": json.dumps({
                            "title"       : str(self.title),
                            "text"        : str(self.text),
                            "actionTitle1": str(self.button[0][0]),
                            "actionURL1"  : str(self.button[0][1]),
                            "actionTitle2": str(self.button[1][0]),
                            "actionURL2"  : str(self.button[1][1]),
                        })
                    }
                elif len(self.button) == 3:
                    self.template = {
                        "msgKey"  : "sampleActionCard3",
                        "msgParam": json.dumps({
                            "title"       : str(self.title),
                            "text"        : str(self.text),
                            "actionTitle1": str(self.button[0][0]),
                            "actionURL1"  : str(self.button[0][1]),
                            "actionTitle2": str(self.button[1][0]),
                            "actionURL2"  : str(self.button[1][1]),
                            "actionTitle3": str(self.button[2][0]),
                            "actionURL3"  : str(self.button[2][1]),
                        })
                    }
                elif len(self.button) == 4:
                    self.template = {
                        "msgKey"  : "sampleActionCard4",
                        "msgParam": json.dumps({
                            "title"       : str(self.title),
                            "text"        : str(self.text),
                            "actionTitle1": str(self.button[0][0]),
                            "actionURL1"  : str(self.button[0][1]),
                            "actionTitle2": str(self.button[1][0]),
                            "actionURL2"  : str(self.button[1][1]),
                            "actionTitle3": str(self.button[2][0]),
                            "actionURL3"  : str(self.button[2][1]),
                            "actionTitle4": str(self.button[3][0]),
                            "actionURL4"  : str(self.button[3][1]),
                        })
                    }
                elif len(self.button) == 5:
                    self.template = {
                        "msgKey"  : "sampleActionCard5",
                        "msgParam": json.dumps({
                            "title"       : str(self.title),
                            "text"        : str(self.text),
                            "actionTitle1": str(self.button[0][0]),
                            "actionURL1"  : str(self.button[0][1]),
                            "actionTitle2": str(self.button[1][0]),
                            "actionURL2"  : str(self.button[1][1]),
                            "actionTitle3": str(self.button[2][0]),
                            "actionURL3"  : str(self.button[2][1]),
                            "actionTitle4": str(self.button[3][0]),
                            "actionURL4"  : str(self.button[3][1]),
                            "actionTitle5": str(self.button[4][0]),
                            "actionURL5"  : str(self.button[4][1]),
                        })
                    }
            else:
                self.template = {
                    "msgKey"  : "sampleActionCard6",
                    "msgParam": json.dumps({
                        "title"       : str(self.title),
                        "text"        : str(self.text),
                        "buttonTitle1": str(self.button[0][0]),
                        "buttonURL1"  : str(self.button[0][1]),
                        "buttonTitle2": str(self.button[1][0]),
                        "buttonURL2"  : str(self.button[1][1]),
                    })
                }
    
    def __str__(self):
        return "[ActionCard]"


class FeedCard(BaseElement):
    
    def __init__(self, links: list):
        """发送FeedCard消息

        Args:
            links: 链接列表，格式：[[title, msgURL, picURL], ...]
        """
        self.links = links
        self.data = {
            "msgtype" : "feedCard",
            "feedCard": {
                "links": []
            }
        }
        for link in self.links:
            if link[0] and link[1]:
                self.data['feedCard']['links'].append({
                    "title" : str(link[0]), "messageURL": _link_detect(str(link[1])),
                    "picURL": _link_detect(str(link[2]))
                })
    
    def __str__(self):
        return "[FeedCard]"


class At:
    
    def __init__(self, target: Union[int, Member], display: str = ""):
        if type(target) == Member:
            self.id = (int(hashlib.sha1(str(target.id)[str(target.id).rfind("$"):].encode('utf-8')).hexdigest(), 16)) % \
                      (10 ** 10) + 1000
            self.target = str(target.staffid)
        else:
            self.id = (int(hashlib.sha1(str(target)[str(target).rfind("$"):].encode('utf-8')).hexdigest(), 16)) % \
                      (10 ** 10) + 1000
            self.target = target
        self.display = display if display else self.target
        self.data = {"atUserIds": [str(self.target)]} if len(str(self.target)) != 11 else {
            "atMobiles": [str(self.target)]}
    
    def __str__(self):
        return "@" + str(self.display)
    
    def __add__(self, other):
        to = copy.deepcopy(self)
        for i in other.data["atUserIds"]:
            to.data["atUserIds"].append(i)
        return to


def _link_detect(link: str) -> str:
    if link:
        if not str(link).startswith("http"):
            link = "https://" + str(link)
    return link



class ConfigError(Exception):
    pass


class GroupSecureKeyError(Exception):
    pass


class UploadFileError(Exception):
    pass


class UploadFileSizeError(Exception):
    pass


class SystemError(Exception):
    pass



import sqlite3


class Cache:
    db: sqlite3.Connection = None
    cursor: sqlite3.Cursor = None

    def __init__(self):
        pass

    @classmethod
    def connect(cls, databaseName: str = "Dingraia_cache.db"):
        cls.db = sqlite3.connect(databaseName)
        cls.cursor = cls.db.cursor()
        
    @classmethod
    def execute(cls, command, result=False):
        cls.cursor.execute(command)
        if result:
            return cls.cursor.fetchall()
        
    @classmethod
    def get_tables(cls):
        res = cls.execute("SELECT name FROM sqlite_master WHERE type='table' AND name NOT LIKE 'sqlite_%';", True)
        return [x[0] for x in res]
    
    @classmethod
    def get_table(cls, table):
        return cls.execute(f"SELECT * FROM {table};", True)
    
    @classmethod
    def create_table(cls, table_name, keys):
        def type_transformer(typ: type):
            if typ == str:
                return "TEXT"
            elif typ == int:
                return "BIGINT"
            return typ
        if keys:
            if isinstance(keys, list):
                if isinstance(keys[0], list) and len(keys[0]) == 2:
                    n_keys = {}
                    for i in keys:
                        n_keys[i[0]] = i[1]
                    keys = n_keys
            elif isinstance(keys, dict):
                pass
            else:
                return ValueError("Keys can only be list or dict, but %s given" % type(keys).__name__)
            types = []
            for k, v in keys.items():
                types += [f"{k} {type_transformer(v)} not null"]
            types = ', '.join(types)
            cls.execute(f"CREATE TABLE {table_name}({types});")
            cls.db.commit()
        else:
            raise ValueError("Keys is empty!")
    
    @classmethod
    def drop_table(cls, table):
        cls.execute(f"DROP TABLE {table};")
        cls.db.commit()
        
    @classmethod
    def rename_table(cls, table, new_name):
        cls.execute(f"ALTER TABLE {table} RENAME TO {new_name};")
        cls.db.commit()
        
    @classmethod
    def add_column(cls, table, key, typ):
        cls.execute(f"ALTER TABLE {table} ADD COLUMN {key} {typ};")
        cls.db.commit()
    
    @classmethod
    def init_tables(cls):
        to_tables = {
            "webhooks": {
                "id": str,
                "url": str
            },
            "group_info": {
                "chatId"            : str,
                "openConversationId": str,
                "info"              : str
            }
        }
        tables = cls.get_tables()
        for t, v in to_tables.items():
            if t not in tables:
                cls.create_table(t, v)

    @classmethod
    def close(cls):
        cls.cursor.close()
        cls.db.close()



from typing import Any

from typing_extensions import Self

from .element import At


class MessageChain:
    
    def __init__(self, *elements, at: list = None):
        self.mes = [s for s in list(elements)]
        self.display = ''.join([str(x) for x in self.mes])
        if at is not None:
            self.mes += [At(at_id) for at_id in at]
    
    def include(self, __obj: type) -> Self:
        return self.get(__obj)  # __class__([i for i in self.mes if isinstance(i, __obj)])
    
    def get_first(self, __obj: type) -> Self:
        return self.get(__obj)[0]
    
    def get(self, element_class: type, count: int = -1) -> list:
        """
        获取消息链中所有特定类型的消息元素

        Args:
            element_class (type): 指定的消息元素的类型, 例如 "Plain", "At", "Image" 等.
            count (int, optional): 至多获取的元素个数

        Return
            list[E]: 获取到的符合要求的所有消息元素; 另: 可能是空列表([]).
        """
        if count == -1:
            count = len(self.mes)
        return [i for i in self.mes if isinstance(i, element_class)][:count]
    
    def __str__(self):
        return self.display
    
    def __getitem__(self, item: type) -> Any:
        if issubclass(item, type):
            return self.get(item)
        else:
            raise NotImplementedError("{0} 不能使用 getitem 方法".format(type(item)))
    
    def __add__(self, other):
        self.mes += other.mes
        self.display = ''.join([str(x) for x in self.mes])
        return self
    
    def __len__(self) -> int:
        return len(self.mes)



import sys
from contextlib import contextmanager
from typing import Union, Callable
import os
from loguru import logger

from .channel import Channel
from .context import channel_instance, saya_instance


class Saya:
    channels = {}
    mirrors = {}
    
    def __init__(self) -> None:
        pass
    
    def set_channel(self):
        saya_instance.set(self)
    
    @contextmanager
    def module_context(self):
        saya_token = saya_instance.set(self)
        yield
        saya_instance.reset(saya_token)
    
    @classmethod
    def current(cls) -> "Saya":
        try:
            return saya_instance.get()
        except LookupError:
            cls().set_channel()
            return saya_instance.get()
    
    def require(self, module_name: str):
        if os.getenv('NoImportModule'):
            return
        if module_name.endswith('.py'):
            module_name = module_name[:-3]
        logger.debug(f"正在载入模块 {module_name}")
        module = __import__(module_name)
        self.channels[module_name] = module
        self.mirrors[module] = module_name
        logger.info(f"模块 {module_name} 载入完成")
    
    def uninstall_channel(self, module_name: Union[str, Callable]):
        if module_name in self.channels or module_name in self.channels.values():
            if module_name in self.mirrors:
                module_name = self.mirrors[module_name]
                del self.mirrors[module_name]
            if module_name in self.channels:
                del self.channels[module_name]
            channel = Channel.current()
            reged = channel.reg_event
            for event in reged.keys():
                module = reged[event]
                modules = list(module.keys())
                if module_name in modules:
                    channel.reg_event[event].pop(module_name)
                    if module_name in sys.modules:
                        del sys.modules[module_name]
        else:
            raise KeyError(f"模块 {module_name} 没有被载入！")



import json
from typing import Union, Any

import aiohttp
import requests
from aiohttp import ClientResponse

from .element import AccessToken

get_token_url = "https://oapi.dingtalk.com/gettoken"


async def url_res(url, method: str = 'GET', headers=None, res: str = 'str', **kwargs) -> Union[ClientResponse, str, dict]:
    async with aiohttp.ClientSession() as session:
        async with session.request(method.upper(), url, headers=headers, **kwargs) as _res:
            resp = await _res.text()
    if res == 'json':
        return json.loads(resp)
    elif res == 'raw':
        return _res
    else:
        return resp


def get_token(app_key, app_secret) -> AccessToken:
    _url = get_token_url + f"?appkey={app_key}&appsecret={app_secret}"
    token = requests.get(_url).json()
    if not token['errcode']:
        return AccessToken(token['access_token'], token['expires_in'])
    else:
        raise ValueError(f"获取Access-Token失败！错误代码：{token['errcode']}，错误信息：{token['errmsg']}")



import asyncio
import base64
import hmac
import importlib.metadata
import json
import signal
import socket
import sys
import urllib.parse
import uuid
from functools import reduce
from typing import List, Dict, Coroutine, Literal
import time
import aiohttp
from aiohttp import ClientSession
import flask
import requests
import websockets
from loguru import logger
import collections
from .VERSION import VERSION
from .callback_handler import callback_handler
from .config import Config, Stream
from .element import *
from .event import MessageEvent
from .event.message import *
from .exceptions import *
from .message.chain import MessageChain
from .message.element import *
from .model import Group, Webhook
from .saya import Channel, Saya
from .signer import sign_js, decrypt
from .tools.debug import delog
from .verify import get_token
from .verify import url_res

send_url = "https://oapi.dingtalk.com/robot/send?access_token={}&timestamp={}&sign={}"
channel = Channel.current()
DINGRAIA_ASCII = r"""
 ____  _                       _
|  _ \(_)_ __   __ _ _ __ __ _(_) __ _
| | | | | '_ \ / _` | '__/ _` | |/ _` |
| |_| | | | | | (_| | | | (_| | | (_| |
|____/|_|_| |_|\__, |_|  \__,_|_|\__,_|
               |___/
"""
ANNOUNCEMENT = "See https://dingraia.gitbook.io/dingraia for documents"

try:
    is_debug = os.path.exists("dingraia_debug.mode")
except:
    is_debug = False


def set_num():
    i = 0
    while True:
        i += 1
        yield i


_no = set_num()


class Dingtalk:
    config: Config = None
    loop: asyncio.AbstractEventLoop = None
    _access_token: AccessToken = AccessToken()
    async_tasks = []
    stream_checker = {}
    """用于检测重复的回调, 键为任务名, 值为容纳50个StreamID的列表"""
    
    def __init__(self, config: Config = None):
        if Dingtalk.config is None:
            Dingtalk.config = config
    
    async def send_message(self, target: Union[Group, Member, OpenConversationId, str, Webhook, None], msg, headers=None):
        """发送普通的文本信息
        
        Args:
            target: 要发送的地址，可以是Group, OpenConversationId, str格式的链接, 或者None发送到测试群
            msg: 要发送的文本
            headers: 要包含的请求头

        Returns:
            List(bool)

        """
        # TODO
        if headers is None:
            headers = {}
        send_data = {}
        response = Response()
        response.recallType = f"Unsupported recall target {type(target).__name__}"
        if isinstance(target, str):
            if target.startswith('cid'):
                target = OpenConversationId(target)
        if isinstance(msg, BaseElement):
            if not isinstance(target, OpenConversationId) and not isinstance(target, Member):
                send_data = msg.data
            else:
                send_data = msg.template
        elif isinstance(msg, File):
            if not msg.mediaId:
                res = await self.upload_file(msg)
                send_data = msg.template
                send_data['media_id'] = res.mediaId
            else:
                send_data['media_id'] = msg.mediaId
            if isinstance(target, Group):
                target = target.conversationId
        else:
            if isinstance(target, OpenConversationId) or isinstance(target, Member):
                send_data = {
                    'msgKey'  : "sampleText",
                    'msgParam': json.dumps({
                        'content': str(msg)
                    })
                }
            else:
                send_data = {
                    "msgtype": "text",
                    "text"   : {
                        "content": str(msg)
                    }
                }
        if isinstance(target, OpenConversationId):
            send_data['robotCode'] = self.config.bot.robotCode
            send_data['openConversationId'] = target.openConversationId
            headers['x-acs-dingtalk-access-token'] = self.access_token
        if isinstance(msg, MessageChain):
            if ats := msg.include(At):
                at = reduce(lambda x, y: x + y, ats)
                send_data["at"] = at.data
        if not target:
            if not self.config.bot.GroupWebhookSecureKey or not self.config.bot.GroupWebhookAccessToken:
                raise ConfigError("Not GroupWebhookSecureKey or GroupWebhookAccessToken provided!")
        if target is None:
            sign = self.get_sign()
            url = send_url.format(sign[2], sign[1], sign[0])
            self.log.info(f"[SEND] <- {repr(str(msg))[1:-1]}")
        elif isinstance(target, Group):
            if time.time() < target.webhook.expired_time:
                url = target.webhook.url
                self.log.info(f"[SEND][{target.name}({int(target)})] <- {repr(str(msg))[1:-1]}")
            else:
                logger.error("群组的Webhook链接已经过期！请检查来源IP是否正确或服务器时钟是否正确")
                return [False, -1]
        elif isinstance(target, Member):
            url = f"https://api.dingtalk.com/v1.0/robot/oToMessages/batchSend"
            headers['x-acs-dingtalk-access-token'] = self.access_token
            self.log.info(f"[SEND][{target.name}({int(target)})] <- {repr(str(msg))[1:-1]}")
            send_data['userIds'] = [target.staffid]
            send_data['robotCode'] = self.config.bot.robotCode
            response.recallType = "personal"
        elif isinstance(target, Webhook):
            if time.time() < target.expired_time:
                url = target.url
            else:
                logger.error("指定的临时Webhook链接已经过期")
                return
        elif isinstance(target, OpenConversationId):
            url = 'https://api.dingtalk.com/v1.0/robot/groupMessages/send'
            response.recallType = "group"
            response.recallOpenConversationId = target
            self.log.info(f"[SEND][{target.name}({int(target)})] <- {repr(str(msg))[1:-1]}")
        else:
            url = str(target)
            self.log.info(f"[SEND] <- {repr(str(msg))[1:-1]}")
        delog.info(send_data, no=60)
        if url and "http" not in url:
            logger.error(f"Wrong send url [{url}]!")
            response.ok = False
            response.text = ""
            response.url = url
            response.recall_type = "Not completed request"
            return response
        response.sendData = send_data
        try:
            resp = await url_res(url, 'POST', json=send_data, headers=headers, res='raw')
            response.ok = resp.ok
            response.text = await resp.text()
            response.url = url
            delog.success("发送完成")
        except Exception as err:
            logger.exception(f"发送失败！", err)
            response.ok = False
            response.text = ""
            response.url = url
            response.recall_type = "Not completed request"
            return response
        else:
            delog.info(response.json, no=40)
            if not response.ok:
                logger.error(f"Failed to send the message!Response: {response.json}")
            else:
                delog.success(f"Success!", no=40)
            return response
    
    async def recall_message(
            self,
            message: Response = None,
            openConversationId: Union[OpenConversationId, Group, str] = None,
            processQueryKeys: Union[str, List[str]] = None,
            robotCode: str = None,
            access_token: str = None,
            inThreadTime: int = 0
    ):
        """撤回一条消息
        
        Args:
            message: 通过send_message发送消息返回的对象
            openConversationId: 群对话ID, 可以是OpenConversationId或Group对象
            processQueryKeys: 消息的加密ID
            robotCode: 机器人的机器码
            access_token: 企业的AccessToken
            inThreadTime: 是否不等待撤回

        Returns:

        """
        async def _run(_message, _openConversationId, _processQueryKeys, _robotCode, _access_token, _inThreadTime):
            _access_token = _access_token or self.access_token
            if message:
                _processQueryKeys = message.json['processQueryKey']
                _openConversationId = message.recallOpenConversationId
            _processQueryKeys = [str(x) for x in
                                (_processQueryKeys if isinstance(_processQueryKeys, list) else [_processQueryKeys])]
            if isinstance(openConversationId, Group):
                _openConversationId = _openConversationId.conversationId
            if _openConversationId is not None:
                _openConversationId = str(_openConversationId)
            if _robotCode is None:
                _robotCode = self.config.bot.robotCode
            post_data = {
                "processQueryKeys"  : _processQueryKeys,
                "robotCode"         : _robotCode
            }
            url = f"https://api.dingtalk.com/v1.0/robot/otoMessages/batchRecall"
            if _openConversationId:
                url = f"https://api.dingtalk.com/v1.0/robot/groupMessages/recall"
                post_data['openConversationId'] = _openConversationId
            await asyncio.sleep(_inThreadTime)
            res = await url_res(url, 'POST', headers={'x-acs-dingtalk-access-token': _access_token},
                                json=post_data, res='json')
            return res
        if not inThreadTime:
            return await _run(message, openConversationId, processQueryKeys, robotCode, access_token, inThreadTime)
        else:
            return self.loop.create_task(_run(message, openConversationId, processQueryKeys, robotCode, access_token, inThreadTime))
    
    async def create_group(self,
                           name,
                           templateId,
                           ownerUserId,
                           icon,
                           userIds: list = None,
                           subAdminIds: list = None,
                           showHistory=False,
                           validation=True,
                           searchable=False,
                           UUID: str = str(uuid.uuid1()),
                           access_token: str = None
                           ):
        userIds = userIds or [ownerUserId]
        userIds = [str(x) for x in userIds]
        userIds = ",".join(userIds)
        subAdminIds = subAdminIds or [ownerUserId]
        subAdminIds = [str(x) for x in subAdminIds]
        subAdminIds = ",".join(subAdminIds)
        access_token = access_token or self.access_token
        url = f"https://oapi.dingtalk.com/topapi/im/chat/scenegroup/create?access_token={access_token}"
        data = {
            "title"                          : name,
            "template_id"                    : templateId,
            "owner_user_id"                  : ownerUserId,
            "user_ids"                       : userIds,
            "subadmin_ids"                   : subAdminIds,
            "uuid"                           : UUID,
            "icon"                           : icon,
            "mention_all_authority"          : 1,
            "show_history_type"              : 1 if showHistory else 0,
            "validation_type"                : 1 if validation else 0,
            "searchable"                     : 1 if searchable else 0,
            "chat_banned_type"               : 0,
            "management_type"                : 1,
            "only_admin_can_ding"            : 0,
            "all_members_can_create_mcs_conf": 1,
            "all_members_can_create_calendar": 0,
            "group_email_disabled"           : 0,
            "only_admin_can_set_msg_top"     : 1,
            "add_friend_forbidden"           : 0,
            "group_live_switch"              : 1,
            "members_to_admin_chat"          : 0
        }
        res = await url_res(url,
                            'POST',
                            json=data, res='json')
        if not res['success']:
            logger.error(f"Cannot create the group!Response: {json.dumps(res, ensure_ascii=False, indent=4)}")
        return res
    
    async def get_group(self, openConversationId: Union[OpenConversationId, Group, str], access_token: str = None):
        """
        
        Args:
            openConversationId: 群对话ID, 可以是OpenConversationId或Group对象
            access_token: 企业的AccessToken

        Returns:

        """
        if isinstance(openConversationId, Group):
            openConversationId = openConversationId.conversationId
        else:
            openConversationId = str(openConversationId)
        access_token = access_token or self.access_token
        res = await url_res(f'https://oapi.dingtalk.com/topapi/im/chat/scenegroup/get?access_token={access_token}',
                            'POST',
                            json={'open_conversation_id': openConversationId}, res='json')
        if res['errcode']:
            if res['errcode'] == 4000003:
                logger.error(f"OpenConversationId 对应的群不是由群模板创建的或没有酷应用支持！")
            res['success'] = False
            return res
        users_res = await url_res(f'https://api.dingtalk.com/v1.0/im/sceneGroups/members/batchQuery',
                                  'POST',
                                  json={'openConversationId': openConversationId, "maxResults": 1000},
                                  headers={'x-acs-dingtalk-access-token': access_token}, res='json')
        res: dict = res['result']
        res['user_ids'] = users_res['memberUserIds']
        res['success'] = True
        return res
    
    async def get_depts(self, deptId: str = "1", access_token: str = None):
        """获取部门ID
        
        Args:
            deptId: 部门ID. 根部门为 1
            access_token: 企业的AccessToken

        Returns:

        """
        access_token = access_token or self.access_token
        url = f"https://oapi.dingtalk.com/topapi/v2/department/listsub?access_token={access_token}"
        res = await url_res(url, 'POST', json={'dept_id': deptId}, res='json')
        return res
    
    async def get_user(self, userStaffId: Union[Member, str], language: str = "zh_CN", access_token: str = None):
        """获取用户详细信息
        
        Args:
            userStaffId: 用户的StaffID
            language: 语言. 默认zh-CN

        Returns:

        """
        if isinstance(userStaffId, Member):
            userStaffId = userStaffId.staffid
        userStaffId = str(userStaffId)
        access_token = access_token or self.access_token
        res = await url_res(
            f'https://oapi.dingtalk.com/topapi/v2/user/get?access_token={access_token}',
            'POST',
            json={"language": language, "userid": userStaffId}, res='json')
        return res
    
    async def remove_user(self, userStaffId: Union[Member, str], access_token: str = None):
        """从组织中直接移除用户
        
        Args:
            userStaffId: 用户的StaffID
            access_token: 组织的AccessToken

        Returns:

        """
        if isinstance(userStaffId, Member):
            userStaffId = userStaffId.staffid
        userStaffId = str(userStaffId)
        access_token = access_token or self.access_token
        res = await url_res(
            f'https://oapi.dingtalk.com/topapi/v2/user/delete?access_token={access_token}',
            'POST',
            json={"userid": userStaffId}, res='json')
        return res
    
    async def create_user(
            self,
            name: str,
            mobilePhone: Union[str, int],
            deptIds: Union[str, List[str]],
            userId: str = None,
            hidePhone: bool = None,
            jobNumber: str = None,
            positionName: str = None,
            personalEmail: str = None,
            organizeEmail: str = None,
            organizeEmailType: Literal['base', 'profession'] = None,
            workPlace: str = None,
            remark: str = None,
            extension: str = None,
            isSenior: bool = None,
            managerUserId: str = None,
            loginEmail: str = None
    ):
        """创建一名用户
        
        Args:
            name: 用户真实名称
            mobilePhone: 用户手机号. 企业内必须唯一
            deptIds: 用户所属部门ID
            userId: 用户的唯一userId. 企业内必须唯一, 不填可自动生成
            hidePhone: 是否隐藏该用户手机号
            jobNumber: 用户工号
            positionName: 用户职位名称
            personalEmail: 用户个人邮箱
            organizeEmail: 用户公司邮箱
            organizeEmailType: 用户公司邮箱类型
            workPlace: 工作地点
            remark: 标记
            extension: 额外信息
            isSenior: 是否开启高管模式
            managerUserId: 管理此用户的经理的UserID
            loginEmail: 登录邮箱

        Returns:

        """
        url = f"https://oapi.dingtalk.com/topapi/v2/user/create?access_token={self.access_token}"
        deptIds = ','.join([str(x) for x in (deptIds if isinstance(deptIds, list) else [deptIds])])
        data = {
            "name"        : name,
            "mobile"      : mobilePhone,
            "telephone"   : mobilePhone,
            "dept_id_list": deptIds,
        }
        if userId:
            data['userid'] = userId
        if positionName:
            data['title'] = positionName
        if jobNumber:
            data['job_number'] = jobNumber
        if personalEmail:
            data['email'] = personalEmail
        if organizeEmail:
            data['org_email'] = organizeEmail
        if organizeEmailType:
            data['org_email_type'] = organizeEmailType
        if hidePhone:
            data['hide_mobile'] = hidePhone
        if remark:
            data['remaek'] = remark
        if workPlace:
            data['work_place'] = workPlace
        if extension:
            data['extension'] = extension
        if isSenior:
            data['senior_mode'] = isSenior
        if managerUserId:
            data['manager_userid'] = managerUserId
        if loginEmail:
            data['login_email'] = loginEmail
        res = await url_res(url, 'POST', json=data, res='json')
        return res
    
    async def mirror_group(self, openConversationId: Union[OpenConversationId, Group, str]):
        """复制群信息和群成员到一个新群. 群必须是场景群
        
        Args:
            openConversationId: 群对话ID, 可以是OpenConversationId或Group对象

        Returns:

        """
        if isinstance(openConversationId, Group):
            openConversationId = openConversationId.conversationId
        openConversationId = str(openConversationId)
        raw = await self.get_group(openConversationId)
        if raw['title'].endswith('_mirror'):
            title = raw['title'][:-7]
        else:
            title = raw['title']
        userIds = raw['user_ids']
        adminUserIds = raw['sub_admin_staff_ids']
        if raw['success']:
            res = await self.create_group(title + " - 转生", raw['template_id'], raw['owner_staff_id'],
                                          raw['icon'], raw['owner_staff_id'], raw['owner_staff_id'],
                                          raw['management_options']['show_history_type'],
                                          raw['management_options']['validation_type'],
                                          raw['management_options']['searchable'])
            new_openConversationId = OpenConversationId(res['result']['open_conversation_id'])
            if res['success']:
                times = 0
                while times < 3:
                    res = await self.get_group(new_openConversationId)
                    if res['ok']:
                        break
                    await asyncio.sleep(1)
                else:
                    logger.error("获取群信息失败超过 3 次!")
                    return {'success': False}
                invite_url = res['group_url']
                await self.send_message(OpenConversationId(openConversationId), MessageChain("新群链接: ", invite_url))
                res = await self.add_member(new_openConversationId, userIds)
            if not res['success']:
                logger.error(
                    f"Error while adding the member! Response: {json.dumps(res, indent=4, ensure_ascii=False)}")
            res = await self.set_admin(new_openConversationId, adminUserIds)
            if not res['success']:
                logger.error(
                    f"Error while setting the admin(s)! Response: {json.dumps(res, indent=4, ensure_ascii=False)}")
        else:
            res = raw
        return res
    
    async def kick_member(self, openConversationId: Union[OpenConversationId, Group, str],
                          memberStaffIds: Union[Member, str, List[Union[Member, str]]]):
        """从群组中踢出一名成员. 群组必须是场景群
        
        Args:
            openConversationId: 群对话ID, 可以是OpenConversationId或Group对象
            memberStaffIds: 成员的StaffID

        Returns:

        """
        if isinstance(openConversationId, Group):
            openConversationId = openConversationId.conversationId
        openConversationId = str(openConversationId)
        if isinstance(memberStaffIds, list):
            if isinstance(memberStaffIds[0], Member):
                memberStaffIds = [x.staffid for x in memberStaffIds]
        else:
            memberStaffIds = [memberStaffIds]
        memberStaffIds = [str(x) for x in memberStaffIds]
        memberStaffIds = ','.join(memberStaffIds)
        res = await url_res(
            f'https://oapi.dingtalk.com/topapi/im/chat/scenegroup/member/delete?access_token={self.access_token}',
            'POST',
            json={"open_conversation_id": openConversationId, "user_ids": memberStaffIds}, res='json')
        return res
    
    async def add_member(self, openConversationId: Union[OpenConversationId, Group, str],
                         memberStaffIds: Union[Member, str, List[Union[Member, str]]]):
        """添加一个成员到群组. 群组必须是场景群
        
        Args:
            openConversationId: 群对话ID, 可以是OpenConversationId或Group对象
            memberStaffIds: 成员的StaffID

        Returns:

        """
        if isinstance(openConversationId, Group):
            openConversationId = openConversationId.conversationId
        openConversationId = str(openConversationId)
        if isinstance(memberStaffIds, list):
            if isinstance(memberStaffIds[0], Member):
                memberStaffIds = [x.staffid for x in memberStaffIds]
        else:
            memberStaffIds = [memberStaffIds]
        memberStaffIds = [str(x) for x in memberStaffIds]
        memberStaffIds = ','.join(memberStaffIds)
        res = await url_res(
            f'https://oapi.dingtalk.com/topapi/im/chat/scenegroup/member/add?access_token={self.access_token}',
            'POST',
            json={"open_conversation_id": openConversationId, "user_ids": memberStaffIds}, res='json')
        return res
    
    async def set_admin(self, openConversationId: Union[OpenConversationId, Group, str],
                        memberStaffIds: Union[Member, str, List[Union[Member, str]]], set_admin: bool = True):
        """设置一个成员是否为管理员. 群组必须是场景群
        
        Args:
            openConversationId: 群对话ID, 可以是OpenConversationId或Group对象
            memberStaffIds: 成员的StaffID
            set_admin: 是否设置为群组管理员

        Returns:

        """
        if isinstance(openConversationId, Group):
            openConversationId = openConversationId.conversationId
        openConversationId = str(openConversationId)
        if isinstance(memberStaffIds, list):
            if isinstance(memberStaffIds[0], Member):
                memberStaffIds = [x.staffid for x in memberStaffIds]
        else:
            memberStaffIds = [memberStaffIds]
        memberStaffIds = [str(x) for x in memberStaffIds]
        res = await url_res(
            f'https://api.dingtalk.com/v1.0/im/sceneGroups/subAdmins',
            'PUT',
            headers={'x-acs-dingtalk-access-token': self.access_token},
            json={"openConversationId": openConversationId, "userIds": memberStaffIds, 'role': 2 if set_admin else 3},
            res='json')
        return res
    
    async def mute_member(
            self,
            openConversationId: Union[OpenConversationId, Group, str],
            memberStaffIds: Union[Member, str, List[Union[Member, str]]],
            muteTime: int = 60
    ):
        """禁言一个成员
        
        Args:
            openConversationId: 群对话ID, 可以是OpenConversationId或Group对象
            memberStaffIds: 成员的StaffID
            muteTime: 禁言时长, 单位为秒, 值为0则解除禁言

        Returns:

        """
        url = f"https://api.dingtalk.com/v1.0/im/sceneGroups/muteMembers/set"
        memberStaffIds = [str(x) for x in (memberStaffIds if isinstance(memberStaffIds, list) else [memberStaffIds])]
        if isinstance(openConversationId, Group):
            openConversationId = openConversationId.conversationId
        openConversationId = str(openConversationId)
        if isinstance(memberStaffIds, list):
            if isinstance(memberStaffIds[0], Member):
                memberStaffIds = [x.staffid for x in memberStaffIds]
        else:
            memberStaffIds = [memberStaffIds]
        res = await url_res(url, 'POST', headers={'x-acs-dingtalk-access-token': self.access_token},
                            json={
                                "openConversationId": openConversationId,
                                "userIdList"        : memberStaffIds,
                                "muteStatus"        : 1 if muteTime else 0,
                                "muteDuration"      : muteTime * 1000
                            }, res='json')
        return res
    
    async def unmute_member(
            self,
            openConversationId: Union[OpenConversationId, Group, str],
            memberStaffIds: Union[Member, str, List[Union[Member, str]]]
    ):
        return await self.mute_member(openConversationId, memberStaffIds, 0)
    
    async def set_off_duty_prompt(
            self,
            text: str = "人家今天下班了呢~请晚些再来找我吧",
            title: str = "钉钉Stream机器人",
            logo: Union[File, str] = "@lALPDfJ6V_FPDmvNAfTNAfQ",
            robotCode: str = None,
            access_token: str = None
    ):
        access_token = access_token or self.access_token
        url = "https://api.dingtalk.com/v1.0/innerApi/robot/stream/away/template/update"
        if isinstance(logo, File):
            if not logo.mediaId:
                logo = await self.upload_file(logo)
            logo = logo.mediaId
        if isinstance(logo, str):
            if not logo.startswith("@"):
                raise ValueError(f"logo {logo} is not a valid value!")
        else:
            raise ValueError(f"logo {logo} is not a valid value!")
        card_data = {
            "config": {
                "autoLayout": True,
                "enableForward": True
            },
            "header": {
                "title": {
                    "type": "text",
                    "text": title
                },
                "logo": logo
            },
            "contents": [
                {
                    "type": "markdown",
                    "text": text,
                    "id": "markdown_" + str(int(time.time()*1000))
                },
                {
                    "type": "divider",
                    "id": "divider_" + str(int(time.time()*1000))
                }
            ]
        }
        data = {
            "robotCode": robotCode or self.config.bot.robotCode,
            "cardData": json.dumps(card_data),
            "cardTemplateId": "StandardCard"
        }
        logger.debug(data)
        res = await url_res(
            url, "POST",
            headers={'x-acs-dingtalk-access-token': access_token},
            json=data
        )
        return res
    
    @staticmethod
    class log:
        def info(*mes):
            logger.info(*mes)
        
        def debug(*mes):
            logger.debug(*mes)
        
        def warning(*mes):
            logger.warning(*mes)
        
        def success(*mes):
            logger.success(*mes)
    
    async def upload_file(self, file: Union[Path, str, File], access_token: str = None) -> File:
        """上传一个文件到钉钉并获取mediaId
        
        Args:
            file: 需要上传的文件
            access_token: 企业的AccessToken

        Returns:
            File

        """
        if not access_token:
            access_token = self.access_token
        if not isinstance(file, File):
            file_type = str(file)
            file_format = file_type[file_type.rfind('.') + 1:]
            if file_format in ['jpg', 'gif', 'png', 'bmp']:
                file_type = 'image'
                res = Image()
            elif file_format in ['amr', 'mp3', 'wav']:
                file_type = 'voice'
                res = Audio()
            elif file_format in ['doc', 'docx', 'xls', 'xlsx', 'ppt', 'pptx', 'zip', 'pdf', 'rar']:
                file_type = 'file'
                res = File()
            elif file_format == 'mp4':
                file_type = 'video'
                res = Video()
            else:
                raise UploadFileError("Cannot upload the file which is not supported!")
            f = open(file, 'rb')
            f.seek(0, os.SEEK_END)
            size = f.tell()
            f.seek(0)
        else:
            size = file.size
            file_type = file.fileType
            f = file.file
            res = file
        if file_type == "image":
            if size > 20 * 1024 * 1024:
                raise UploadFileSizeError("Image file is limited under 20M, but %sM given!" % (size / (1024 ** 2)))
        elif file_type == "voice":
            if size > 2 * 1024 * 1024:
                raise UploadFileSizeError("Voice file is limited under 2M, but %sM given!" % (size / (1024 ** 2)))
        elif file_type == "video":
            if size > 20 * 1024 * 1024:
                raise UploadFileSizeError("Video file is limited under 20M, but %sM given!" % (size / (1024 ** 2)))
        else:
            if size > 20 * 1024 * 1024:
                raise UploadFileSizeError("Normal file is limited under 20M, but %sM given!" % (size / (1024 ** 2)))
        if not access_token:
            access_token = self.access_token
        async with aiohttp.ClientSession() as session:
            async with session.post(f'https://oapi.dingtalk.com/media/upload?access_token={access_token}',
                                    data={'type': file_type, 'media': f}) as resp:
                resp = await resp.json()
        res.mediaId = resp['media_id']
        return res
    
    @property
    def access_token(self):
        """当前企业的AccessToken, 会在调用时自动更新"""
        if self._access_token.ok:
            return self._access_token.token
        else:
            self._access_token = get_token(self.config.bot.appKey, self.config.bot.appSecret)
            return self._access_token.token
    
    channel = Channel.current()
    callbacks = []
    
    @logger.catch
    async def bcc(self, data: dict):
        delog.info(json.dumps(data, indent=2, ensure_ascii=False), no=50)
        _e = self.disPackage(data)
        if _e.get('success'):
            _e['send_data'].append(self)
            # if isinstance(_e.get('event_type'), BasicMessage):
            #     self.log.info(_e.get('send_data'))
            if not isinstance(_e.get('event_type'), list):
                _e['event_type'] = [_e.get('event_type')]
            for event in _e.get('event_type'):
                if event is not None:
                    await channel.radio(event, *_e.get('send_data'))
        if not _e:
            logger.warning("无法解包！")
            return ""
        return _e.get('returns') or {'err': 0}
    
    @logger.catch
    def disPackage(self, data: dict) -> dict:
        if "conversationType" in data:
            conversationType = data.get("conversationType")
            if conversationType is not None:
                bot = Bot(origin=data)
                group = Group(origin=data)
                member = Member(origin=data)
                if conversationType == "2":
                    at_users = [userid.get("dingtalkId") for userid in data.get("atUsers") if
                                userid.get("dingtalkId")[userid.get("dingtalkId").rfind('$'):] != bot.origin_id]
                else:
                    at_users = []
                # if data.get('msgtype') != 'text':
                #     raise ValueError("不支持解析文本以外的消息")
                mes = data.get('text', {}).get('content')
                out_mes = mes
                for _ in out_mes:
                    if mes.startswith(" "):
                        mes = mes[1:]
                    else:
                        break
                # logger.info(at_users)
                message = MessageChain(mes, at=at_users)
                self.log.info(f"[RECV][{group.name}({group.id})] {member.name}({member.id}) -> {message}")
                event = MessageEvent(data.get('msgtype'), data.get('msgId'), data.get('isInAtList'), message, group,
                                     member)
                return {
                    "success"   : True,
                    "send_data" : [group, member, message, event, bot],
                    "event_type": [GroupMessage],
                    "returns"   : ""
                }
            else:
                logger.error("不支持的对话类型")
                logger.error(json.dumps(data, indent=2, ensure_ascii=False))
                return {
                    "success"   : False,
                    "send_data" : [],
                    "event_type": [None],
                    "returns"   : ""
                }
        elif "encrypt" in data:
            app = Dingtalk()
            if app.config is not None:
                _e = None
                key = app.config.event_callback.AesKey
                try:
                    _e = decrypt(data.get("encrypt"), key)
                except json.JSONDecodeError:
                    logger.error("无法解密，请检查配置的事件回调密钥")
                    return {
                        "success"   : False,
                        "send_data" : [],
                        "event_type": [None],
                        "returns"   : "400 Bad"
                    }
                except Exception as err:
                    logger.exception(err)
                    return {
                        "success"   : False,
                        "send_data" : [],
                        "event_type": [None],
                        "returns"   : "400 Bad"
                    }
                if _e is None:
                    logger.error("无法解密，请检查配置的事件回调密钥")
                    return {
                        "success"   : False,
                        "send_data" : [],
                        "event_type": [None],
                        "returns"   : ""
                    }
                if type(_e) == dict:
                    logger.info(_e)
                    res = callback_handler(_e, data)
                    return_data = {
                        "success"   : True,
                        "send_data" : [res],
                        "event_type": [res],
                        "returns"   : sign_js(
                            AES_KEY=app.config.event_callback.AesKey,
                            Token=app.config.event_callback.Token,
                            CropID=app.config.event_callback.CropId
                        )
                    }
                else:
                    return {
                        "success"   : False,
                        "send_data" : [],
                        "event_type": None,
                        "returns"   : ""
                    }
                # return_data['event_type'].append(BasicEvent(data, _e))
                return return_data
            return {
                "success"   : False,
                "send_data" : [],
                "event_type": [None],
                "returns"   : ""
            }
        elif 'EventType' in data:
            logger.info(data)
            res = callback_handler(data)
            return {
                "success"   : True,
                "send_data" : [res],
                "event_type": [res],
                "returns"   : ""
            }
        else:
            logger.error(f"未知的回调类型:{json.dumps(data, indent=2, ensure_ascii=False)}")
            return {
                "success"   : False,
                "send_data" : [],
                "event_type": [None],
                "returns"   : ""
            }
    
    @classmethod
    def get_sign(cls, secure_key: str = None):
        if secure_key is None:
            if cls.config:
                secure_key = cls.config.bot.GroupWebhookSecureKey
        if not secure_key:
            raise GroupSecureKeyError("Not GroupSecureKey provided!")
        timestamp = str(round(time.time() * 1000))
        sign_str = timestamp + '\n' + secure_key
        sign = hmac.new(secure_key.encode("utf-8"), sign_str.encode("utf-8"), hashlib.sha256).digest()
        sign = base64.b64encode(sign)
        sign = urllib.parse.quote_plus(sign)
        return sign, timestamp, secure_key
    
    @staticmethod
    async def _send(url, send_data, headers=None):
        delog.info(f"发送中:{url}", no=40)
        if url and "http" not in url:
            url = 'http://' + url
        if not url:
            logger.error("Empty send url!")
            return [False]
        try:
            resp = await url_res(url, method='POST', json=send_data, headers=headers, res='json')
            delog.success("发送完成")
        except Exception as err:
            logger.exception(f"发送失败！", err)
            return [False]
        else:
            delog.info(resp, no=40)
            if 'errcode' in resp:
                if not resp['errcode']:
                    delog.success(f"Success!", no=40)
                    return [True, resp]
                else:
                    logger.error(f"Failed to send the message!err_code：{resp['errcode']}, err_msg：{resp['errmsg']}")
                    return [False, resp['errcode'], resp['errmsg']]
            else:
                if 'processQueryKey' in resp:
                    delog.success(f"Success!", no=40)
                    return [True, resp]
                else:
                    logger.error(f"Failed to send the message!err_msg：{resp}, send_data: {send_data}")
                    return [False, resp]
    
    @staticmethod
    def _start_topic():
        
        def get_dist_map() -> Dict[str, str]:
            """获取与项目相关的发行字典"""
            dist_map: Dict[str, str] = {}
            for dist in importlib.metadata.distributions():
                name: str = dist.metadata["Name"]
                version: str = dist.metadata["Version"]
                if not name or not version:
                    continue
                if name.startswith(('dingraia', 'websocket', 'asyncio', 'flask')):
                    dist_map[name] = max(version, dist_map.get(name, ""))
            return dist_map
        
        try:
            x, y = os.get_terminal_size().columns, os.get_terminal_size().lines
        except:
            x, y = 80, 20
        _topic = f"[Dingraia v{VERSION}]"
        _topic_len = len(_topic)
        _topic_split = "=" * (((x - _topic_len) // 2) * 2)
        __topic = _topic_split[:len(_topic_split) // 2] + _topic + _topic_split[len(_topic_split) // 2:]
        ver = ""
        if vers := get_dist_map():
            for k, v in vers.items():
                ver += f"\n<magenta>{k}</>: <blue>{v}</>"
        announcement = DEBUG = ""
        if ANNOUNCEMENT:
            announcement = "Announcements:\n" + ANNOUNCEMENT
        if is_debug:
            DEBUG = "<yellow>Warning</>: <red>Debug mode is on!</>\n"
        logger.opt(colors=True).info("\n" * 2 + __topic + DINGRAIA_ASCII + f"{ver}\n\n" + announcement + "\n" + DEBUG)
        logger.info("Preparing loading...")
    
    def start(self, flask_app: flask.Flask = None, **kwargs):
        """
        
        Args:
            flask_app: 你的 Flask app 对象
            **kwargs: 传递给 Flask 的参数

        Returns:
            None

        """
        Channel().set_channel()
        Saya().set_channel()
        self._start_topic()
        if self.config.stream:
            logger.info(f"Loading {len(self.config.stream)} stream task{'s' if len(self.config.stream) > 1 else ''}")
            for stream in self.config.stream:
                self._create_stream(stream)
            self.loop.create_task(self.stop(True))
        if flask_app:
            from flask import request, jsonify
            
            @logger.catch
            @flask_app.route('/', methods=["POST"])
            async def receive_data():
                res = await self.bcc(request.json)
                if res:
                    if isinstance(res, dict):
                        return jsonify(res)
                    else:
                        return res
            
            flask_app.run(**kwargs)
        else:
            self.loop.run_forever()
    
    def _create_stream(self, stream: Stream):
        """创建并开始一个异步Stream任务. 不推荐自行调用
        
        Args:
            stream: Stream对象

        Returns:
            None

        """
        
        WS_CONNECT_URL = "https://api.dingtalk.com/v1.0/gateway/connections/open"
        
        def get_host_ip():
            ip = ""
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                s.connect(('8.8.8.8', 80))
                ip = s.getsockname()[0]
            finally:
                s.close()
                return ip
        
        async def open_connection(task_name: str):
            logger.info(f'[{task_name}] Requesting stream')
            request_headers = {
                'Content-Type': 'application/json',
                'Accept'      : 'application/json',
                'User-Agent'  : f'Dingraia/{VERSION} (+https://github.com/MeiHuaGuangShuo/Dingraia)'
            }
            request_body = json.dumps({
                'clientId'     : stream.AppKey,
                'clientSecret' : stream.AppSecret,
                'subscriptions': [
                    {'type': 'EVENT', 'topic': '*'},
                    {'type': 'CALLBACK', 'topic': '/v1.0/im/bot/messages/get'},
                    {'type': 'CALLBACK', 'topic': '/v1.0/im/bot/messages/delegate'},
                    {'type': 'CALLBACK', 'topic': '/v1.0/card/instances/callback'},
                ],
                'ua'           : f'Dingraia/{VERSION}',
                'localIp'      : get_host_ip()
            }).encode('utf-8')
            response = requests.post(WS_CONNECT_URL,
                                     headers=request_headers,
                                     data=request_body)
            http_body = response.json()
            if not response.ok:
                logger.error(f"[{task_name}] Open connection failed, Reason: {response.reason}, Response: {http_body}")
                if response.status_code == 401:
                    logger.warning(f"[{task_name}] The AppKey or AppSecret maybe inaccurate")
                    await self.stop()
                    
                    return False
                return None
            return response.json()
        
        async def route_message(json_message, websocket: websockets.WebSocketServer, task_name: str):
            result = ''
            try:
                msg_type = json_message.get('type', '')
                headers = json_message.get('headers', {})
                data = json.loads(json_message.get('data', {}))
                topic = headers.get('topic', '')
                if task_name not in self.stream_checker:
                    self.stream_checker[task_name] = collections.deque([], 50)
                    self.stream_checker[task_name].append(headers.get('eventId', str(time.time())))
                elif headers.get('eventId', str(time.time())) in self.stream_checker[task_name]:
                    if is_debug:
                        logger.warning(f"Same Callback. ID:{headers.get('eventId', str(time.time()))}")
                    return result
                else:
                    self.stream_checker[task_name].append(headers.get('eventId', str(time.time())))
                if msg_type == 'SYSTEM':
                    if topic == 'disconnect':
                        result = 'disconnect'
                        logger.warning(
                            f"[{task_name}] [System] Client was offered to disconnect, message: {json_message}")
                    else:
                        logger.info(f"[{task_name}] [System] topic: {topic}")
                    headers['topic'] = "pong"
                    await websocket.send(json.dumps({
                        'code'   : 200,
                        'headers': headers,
                        'message': 'OK',
                        'data'   : json_message['data'],
                    }))
                else:
                    logger.debug(f"[{task_name}] " + json.dumps(json_message, indent=4, ensure_ascii=False))
                    if 'eventType' in headers:
                        data['EventType'] = headers['eventType']
                    data['CropId'] = headers.get('eventCropId')
                    await self.bcc(data)
                    await websocket.send(json.dumps({
                        'code'   : 200,
                        'headers': headers,
                        'message': 'OK',
                        'data'   : json_message['data'],
                    }))
            except Exception as err:
                logger.exception(f"[{task_name}] Error happened while handing the message", err)
            return result
        
        async def main_stream(task_name: str):
            while True:
                connection = await open_connection(task_name)
                
                if not connection:
                    if connection is None:
                        logger.error(f'[{task_name}] Open websocket connection failed')
                        logger.warning(f"[{task_name}] Websocket Connection will be reconnected after 5 seconds")
                        time.sleep(5)
                        logger.warning(f"[{task_name}] Reconnecting...")
                        continue
                    if not connection:
                        logger.error(f'[{task_name}] Request connection failed!')
                    return
                
                uri = '%s?ticket=%s' % (connection['endpoint'], urllib.parse.quote_plus(connection['ticket']))
                async with websockets.connect(uri) as websocket:
                    logger.success(f"[{task_name}] Websocket connected")
                    async for raw_message in websocket:
                        json_message = json.loads(raw_message)
                        route_result = await route_message(json_message, websocket, task_name)
                        if route_result == "disconnect":
                            break
                    # self.websocket.close()
            return
        
        try:
            self.loop = asyncio.get_event_loop()
            no = next(_no)
            name = f"#{no} Main Stream"
            self.async_tasks.append(self.loop.create_task(main_stream(name), name=name))
            logger.info(f"Create Stream Task - {name}")
        except asyncio.exceptions.CancelledError:
            logger.warning("Program forced to be exit!")
            sys.exit(0)
    
    def create_task(self, coroutine: Coroutine, name: str = "Task"):
        """创建一个异步任务, 此任务会在stop函数被调用时取消
        
        Args:
            coroutine: 要加入的异步任务
            name: 任务名称, 用于标识

        Returns:
            tuple(Task, name)

        """
        task = self.loop.create_task(coroutine)
        task.set_name(name := (f"#{next(_no)} " + name))
        self.async_tasks.append(task)
        logger.info(f"Create async task [{name}]")
        return task, name
    
    async def stop(self, waitForSignal=False):
        if waitForSignal:
            while not exit_signal:
                await asyncio.sleep(0.5)
        else:
            logger.warning("Stop signal was called!")
        self.log.info("Stopping task loop")
        tasks = self.async_tasks
        names = ', '.join([x.get_name() for x in self.async_tasks])
        logger.info(f"Cancelling async tasks: [{names}]")
        if tasks:
            for task in tasks:
                is_cancelled = task.cancel()
                name = task.get_name()
                if task in self.async_tasks:
                    if not is_cancelled:
                        logger.error(f"Task [{name}] canceled failed!")
                    else:
                        logger.success(f"Task [{name}] canceled successfully")
        self.loop.stop()


exit_signal = False


def exit(signum, frame):
    global exit_signal
    logger.warning(f"Received Signal {signum}")
    exit_signal = True


signal.signal(signal.SIGINT, exit)
signal.signal(signal.SIGTERM, exit)



from typing import List


class Bot:
    
    def __init__(self, AppKey: str, AppSecret: str, robotCode: str, GroupWebhookAccessToken: str = None,
                 GroupWebhookSecureKey: str = None):
        self.appKey = AppKey
        self.appSecret = AppSecret
        self.robotCode = robotCode
        self.GroupWebhookAccessToken = GroupWebhookAccessToken
        self.GroupWebhookSecureKey = GroupWebhookSecureKey


class CallBack:
    
    def __init__(self, AesKey: str, Token: str, CropId: str):
        self.AesKey = AesKey
        self.Token = Token
        self.CropId = CropId
        self.elements = [self.AesKey, self.Token, self.CropId]
    
    def __getitem__(self, item):
        return self.elements[item]
    
    def __setitem__(self, key, value):
        self.elements[key] = value


class Stream:
    
    def __init__(self, AppKey: str, AppSecret: str):
        self.AppKey = AppKey
        self.AppSecret = AppSecret


class Config:
    
    def __init__(self,
                 event_callback: CallBack = None,
                 bot: Bot = None,
                 stream: List[Stream] = None
                 ):
        self.event_callback = event_callback
        self.bot = bot
        self.stream = stream
        if not isinstance(self.stream, list):
            self.stream = [self.stream]



import time

from loguru import logger


class TimeCost:
    
    def __init__(self, title=None):
        self.title = title
    
    def __enter__(self):
        self.start_time = time.time()
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        cost_time = time.time() - self.start_time
        logger.info(f"Time{f'({self.title})' if self.title is not None else ''}: {cost_time}s")



from ..message.chain import MessageChain
from ..model import Group, Member


class MessageEvent:
    
    def __init__(self, type: str, id: str, atd: bool, message: MessageChain, group: Group, member: Member):
        self.type = type
        self.id = id
        self.atd = atd
        self.message = message
        self.group = group
        self.member = member



import base64
import binascii
import hashlib
import hmac
import io
import json
import string
import struct
import time
import urllib.parse
from random import choice

from Crypto.Cipher import AES
from loguru import logger

from dingraia.tools.debug import delog


def decrypt(encrypt_data, aes_key):
    key = base64.b64decode(aes_key + '=')
    encrypt_data = base64.b64decode(encrypt_data)
    iv = encrypt_data[:16]
    encrypted_msg = encrypt_data[16:]
    cipher = AES.new(key, AES.MODE_CBC, iv)
    decrypted_data = cipher.decrypt(encrypted_msg)
    msg_len = int.from_bytes(decrypted_data[:4], byteorder="big")
    msg = decrypted_data[4:4 + msg_len]
    return json.loads(msg.decode())


def encrypt(data, Token, AES_KEY, CropID):
    dingCrypto = DingCallbackCrypto3(Token, AES_KEY, CropID)
    return dingCrypto.getEncryptedMap(data)


def sign_js(Token, AES_KEY, CropID):
    return encrypt("success", Token, AES_KEY, CropID)


@logger.catch
def get_sign(secure_key: str):
    timestamp = str(round(time.time() * 1000))
    sign_str = timestamp + '\n' + secure_key
    sign = hmac.new(secure_key.encode("utf-8"), sign_str.encode("utf-8"), hashlib.sha256).digest()
    sign = base64.b64encode(sign)
    sign = urllib.parse.quote_plus(sign)
    delog.success(f"成功加签", no=40)
    return sign, timestamp


class DingCallbackCrypto3:
    def __init__(self, token, encodingAesKey, key):
        self.encodingAesKey = encodingAesKey
        self.key = key
        self.token = token
        self.aesKey = base64.b64decode(self.encodingAesKey + '=')
    
    # 生成回调处理完成后的success加密数据
    def getEncryptedMap(self, content):
        encryptContent = self.encrypt(content)
        timeStamp = str(int(time.time()))
        nonce = self.generateRandomKey(16)
        sign = self.generateSignature(nonce, timeStamp, self.token, encryptContent)
        return {'msg_signature': sign, 'encrypt': encryptContent, 'timeStamp': timeStamp, 'nonce': nonce}
    
    # 解密钉钉发送的数据
    def getDecryptMsg(self, msg_signature, timeStamp, nonce, content):
        """
        解密
        :param msg_signature: 消息体签名
        :param timeStamp: 时间戳
        :param nonce: 随机字符串
        :param content: 钉钉返回的encrypt字段
        :return:
        """
        sign = self.generateSignature(nonce, timeStamp, self.token, content)
        # print(sign, msg_signature)
        if msg_signature != sign:
            raise ValueError('signature check error')
        content = base64.decodebytes(content.encode('UTF-8'))  # 钉钉返回的消息体
        iv = self.aesKey[:16]  # 初始向量
        aesDecode = AES.new(self.aesKey, AES.MODE_CBC, iv)
        decodeRes = aesDecode.decrypt(content)
        pad = int(decodeRes[-1])
        if pad > 32:
            raise ValueError('Input is not padded or padding is corrupt')
        decodeRes = decodeRes[:-pad]
        l = struct.unpack('!i', decodeRes[16:20])[0]
        nl = len(decodeRes)
        if decodeRes[(20 + l):].decode() != self.key:
            raise ValueError(f'corpId 校验错误:{decodeRes[(20 + l):].decode(), self.key}')
        return decodeRes[20:(20 + l)].decode()
    
    def encrypt(self, content):
        """
        加密
        :param content: 要加密的消息体
        :return: 加密完成的字符串
        """
        msg_len = self.length(content)
        content = ''.join([self.generateRandomKey(16), msg_len.decode(), content, self.key])
        contentEncode = self.pks7encode(content).encode()
        iv = self.aesKey[:16]
        aesEncode = AES.new(self.aesKey, AES.MODE_CBC, iv)
        aesEncrypt = aesEncode.encrypt(contentEncode)
        return base64.encodebytes(aesEncrypt).decode('UTF-8')
    
    # 生成回调返回使用的签名值
    @staticmethod
    def generateSignature(nonce, timestamp, token, msg_encrypt):
        # print(type(nonce), type(timestamp), type(token), type(msg_encrypt))
        v = msg_encrypt
        signList = ''.join(sorted([nonce, timestamp, token, v]))
        return hashlib.sha1(signList.encode()).hexdigest()
    
    @staticmethod
    def length(content):
        """
        将msg_len转为符合要求的四位字节长度
        :param content:
        :return:
        """
        l = len(content)
        return struct.pack('>l', l)
    
    @staticmethod
    def pks7encode(content):
        """
        安装 PKCS#7 标准填充字符串
        :param content: str
        :return: str
        """
        l = len(content)
        output = io.StringIO()
        val = 32 - (l % 32)
        for _ in range(val):
            output.write('%02x' % val)
        # print "pks7encode",content,"pks7encode", val, "pks7encode", output.getvalue()
        return content + binascii.unhexlify(output.getvalue()).decode()
    
    @staticmethod
    def pks7decode(content):
        nl = len(content)
        val = int(binascii.hexlify(content[-1]), 16)
        if val > 32:
            raise ValueError('Input is not padded or padding is corrupt')
        
        l = nl - val
        return content[:l]
    
    @staticmethod
    def generateRandomKey(size,
                          chars=string.ascii_letters + string.ascii_lowercase + string.ascii_uppercase + string.digits):
        """
        生成加密所需要的随机字符串
        :param size:
        :param chars:
        :return:
        """
        return ''.join(choice(chars) for i in range(size))



from IPython import embed
import os
from pathlib import Path
from .message.chain import MessageChain
from .message.element import *
from .element import *


path = ''
oapi = "https://oapi.dingtalk.com"
api = "https://api.dingtalk.com"
os.environ['NoImportModule'] = '1'
if not os.path.exists(Path("main.py")):
    while True:
        path = input('Input filename(Empty to exit) > ')
        if not os.path.exists(Path(path)):
            print("File not found!")
path = path or "main.py"
path = path.replace('.py', '')
app = __import__(path).app
embed(colors='Neutral', using='asyncio')




def ColoredFormatter(message: str):
    color_map = {
        "red"    : "\033[1;31m",
        "green"  : "\033[1;32m",
        "yellow" : "\033[1;33m",
        "blue"   : "\033[1;34m",
        "magenta": "\033[1;35m",
        "cyan"   : "\033[1;36m",
        "white"  : "\033[1;37m",
        "reset"  : "\033[0m",
    }
    for color in color_map:
        message = message.replace(f"<{color}>", color_map[color])
        message = message.replace(f"</{color}>", color_map["reset"])
    return message


s = 0
for colour in range(256):
    if s > 16:
        s = 0
        print()
    print(f"\x1b[38;5;{colour}m\u25a0\x1b[0m", end="")
    s += 1



import inspect
import sys
from typing import Any

from loguru import logger

logger.remove()
logger.add(sys.stdout, filter=lambda record: record['extra'].get("name") != "debug_log", enqueue=True)
logger.add(sys.stdout,
           format="<green>{time:YYYY-MM-DD HH:mm:ss:SSS}</green> <red>|</red> <level>{level: <8}</level> "
                  "<red>|</red> <level>{message}</level>",
           filter=lambda record: record['extra'].get("name") == "debug_log",
           enqueue=True)
debug_logger = logger.bind(name="debug_log")


class delog:
    """简易的DEBUG方法，方便调试"""
    
    is_debug: bool = False
    no: int = 60
    current_modules: list = []
    forbidden_modules: list = []
    
    @classmethod
    def info(cls, __message, *args: Any, no: int = 60, **kwargs: Any):
        lies = cls._get_caller(inspect.currentframe())
        access = False
        if cls.current_modules:
            if lies[0] in cls.current_modules:
                access = True
            else:
                return
        if lies[0] in cls.forbidden_modules:
            return
        if cls.is_debug and (no >= cls.no or access):
            debug_logger.opt(colors=True).info(
                f"<cyan>{lies[0]}</><red>:</><cyan>{lies[1]}</><red>:</><cyan>{lies[2]}</> <red>-</> " + str(
                    __message).replace("<", "\\<"), *args, **kwargs)
    
    @classmethod
    def debug(cls, __message, *args: Any, no: int = 60, **kwargs: Any):
        lies = cls._get_caller(inspect.currentframe())
        access = False
        if cls.current_modules:
            if lies[0] in cls.current_modules:
                access = True
            else:
                return
        if lies[0] in cls.forbidden_modules:
            return
        if cls.is_debug and (no >= cls.no or access):
            debug_logger.opt(colors=True).debug(
                f"<cyan>{lies[0]}</><red>:</><cyan>{lies[1]}</><red>:</><cyan>{lies[2]}</> <red>-</> " + str(
                    __message).replace("<", "\\<"), *args, **kwargs)
    
    @classmethod
    def warning(cls, __message, *args: Any, no: int = 60, **kwargs: Any):
        lies = cls._get_caller(inspect.currentframe())
        access = False
        if cls.current_modules:
            if lies[0] in cls.current_modules:
                access = True
            else:
                return
        if lies[0] in cls.forbidden_modules:
            return
        if cls.is_debug and (no >= cls.no or access):
            debug_logger.opt(colors=True).warning(
                f"<cyan>{lies[0]}</><red>:</><cyan>{lies[1]}</><red>:</><cyan>{lies[2]}</> <red>-</> " + str(
                    __message).replace("<", "\\<"), *args, **kwargs)
    
    @classmethod
    def success(cls, __message, *args: Any, no: int = 60, **kwargs: Any):
        lies = cls._get_caller(inspect.currentframe())
        access = False
        if cls.current_modules:
            if lies[0] in cls.current_modules:
                access = True
            else:
                return
        if lies[0] in cls.forbidden_modules:
            return
        if cls.is_debug and (no >= cls.no or access):
            debug_logger.opt(colors=True).success(
                f"<cyan>{lies[0]}</><red>:</><cyan>{lies[1]}</><red>:</><cyan>{lies[2]}</> <red>-</> " + str(
                    __message).replace("<", "\\<"), *args, **kwargs)
    
    @classmethod
    def error(cls, __message, *args: Any, no: int = 60, **kwargs: Any):
        lies = cls._get_caller(inspect.currentframe())
        access = False
        if cls.current_modules:
            if lies[0] in cls.current_modules:
                access = True
            else:
                return
        if lies[0] in cls.forbidden_modules:
            return
        if cls.is_debug and (no >= cls.no or access):
            debug_logger.opt(colors=True).error(
                f"<cyan>{lies[0]}</><red>:</><cyan>{lies[1]}</><red>:</><cyan>{lies[2]}</> <red>-</> " + str(
                    __message).replace("<", "\\<"), *args, **kwargs)
    
    @classmethod
    def exception(cls, __message, *args: Any, no: int = 60, **kwargs: Any):
        lies = cls._get_caller(inspect.currentframe())
        access = False
        if cls.current_modules:
            if lies[0] in cls.current_modules:
                access = True
            else:
                return
        if lies[0] in cls.forbidden_modules:
            return
        if cls.is_debug and (no >= cls.no or access):
            debug_logger.opt(colors=True).exception(
                f"<cyan>{lies[0]}</><red>:</><cyan>{lies[1]}</><red>:</><cyan>{lies[2]}</> <red>-</> " + str(
                    __message).replace("<", "\\<"), *args, **kwargs)
    
    @classmethod
    def start(cls):
        """启用DEBUG日志"""
        cls.is_debug = True
    
    @classmethod
    def stop(cls):
        """停止DEBUG日志"""
        cls.is_debug = False
    
    @classmethod
    def level(cls, no: int):
        """设置记录日志的优先级"""
        cls.no = no
    
    @classmethod
    def add_module(cls, current_module: str) -> bool:
        """添加一个模块，用以只输出该模块"""
        try:
            cls.current_modules.append(current_module)
            return True
        except Exception as err:
            logger.exception(err)
            return False
    
    @classmethod
    def remove_module(cls, current_module: str) -> bool:
        """删除一个模块，使其不再输出"""
        try:
            cls.current_modules.remove(current_module)
            return True
        except Exception as err:
            logger.exception(err)
            return False
    
    @classmethod
    def clear_module(cls) -> bool:
        """清除之前指定的模块，重新恢复按优先级输出日志"""
        try:
            cls.current_modules = []
            return True
        except Exception as err:
            logger.exception(err)
            return False
    
    @classmethod
    def block_module(cls, current_module: str) -> bool:
        """屏蔽一个模块，使其不再输出"""
        try:
            cls.forbidden_modules.append(current_module)
            return True
        except Exception as err:
            logger.exception(err)
            return False
    
    @classmethod
    def remove_block_module(cls, current_module: str) -> bool:
        """取消一个屏蔽的模块，使其不再输出"""
        try:
            cls.forbidden_modules.remove(current_module)
            return True
        except Exception as err:
            logger.exception(err)
            return False
    
    @classmethod
    def clear_block_module(cls) -> bool:
        """清除之前屏蔽的模块"""
        try:
            cls.forbidden_modules = []
            return True
        except Exception as err:
            logger.exception(err)
            return False
    
    @staticmethod
    def _get_caller(f_back) -> list:
        caller_frame = f_back.f_back
        module_name = caller_frame.f_back.f_globals["__name__"]
        line = caller_frame.f_lineno
        func_name = caller_frame.f_code.co_name
        return [module_name.replace("<", "\\<"), func_name.replace("<", "\\<"), line]



import asyncio
import functools
import inspect
import itertools
from concurrent.futures import ThreadPoolExecutor
from typing import Union

from loguru import logger

from .builtins.broadcast.schema import ListenerSchema
from .context import channel_instance


class Channel:
    reg_event = {}
    
    def __init__(self) -> None:
        pass
    
    def use(self, ListenEvent: Union[list, ListenerSchema]):
        if type(ListenEvent) == ListenerSchema:
            ListenEvent = ListenEvent.listening_events
        
        def wrapper(func):
            module_name = inspect.getmodule(func).__name__
            for event in ListenEvent:
                if event in self.reg_event:
                    if module_name in self.reg_event[event]:
                        self.reg_event[event][module_name].append(func)
                    else:
                        self.reg_event[event][module_name] = [func]
                else:
                    self.reg_event[event] = {}
                    self.reg_event[event][module_name] = [func]
            return func
        
        return wrapper
    
    async def radio(self, RadioEvent, *args):
        # logger.debug(f"{type(RadioEvent) in self.reg_event} {RadioEvent} {type(RadioEvent)} {self.reg_event}")
        if type(RadioEvent) != type:
            RadioEvent = type(RadioEvent)
        if RadioEvent in self.reg_event:
            modules = list(itertools.chain(*self.reg_event[RadioEvent].values()))
            async_tasks = []
            loop = asyncio.get_event_loop()
            with ThreadPoolExecutor() as pool:
                for f in modules:
                    send = {}
                    sig = inspect.signature(f)
                    params = sig.parameters
                    for name, param in params.items():
                        for typ in args:
                            if isinstance(typ, param.annotation):
                                send[name] = typ
                    if inspect.iscoroutinefunction(f):
                        async_tasks.append(loop.create_task(logger.catch(f)(**send)))
                    else:
                        async_tasks.append(loop.run_in_executor(pool, functools.partial(logger.catch(f), **send)))
                if async_tasks:
                    await asyncio.gather(*async_tasks)
                    async_tasks.clear()
    
    def set_channel(self):
        channel_instance.set(self)
    
    @classmethod
    def current(cls) -> "Channel":
        try:
            return channel_instance.get()
        except LookupError:
            cls().set_channel()
            return channel_instance.get()



from setuptools import setup, find_packages

setup(
    name='dingraia',
    version='2.0.2',
    packages=find_packages(),
    url='https://github.com/MeiHuaGuangShuo/Dingraia',
    author='MeiHuaGuangShuo',
    author_email='meihuaguangshuo@gmail.com',
    description='A Dingtalk robot framework based on Python, supporting http and stream mode.',
    long_description='A Dingtalk robot framework based on official, easier and advanced to use.',
    install_requires=[
        "loguru",
        "requests",
        "aiohttp",
        "pycryptodome",
        "websockets"
    ],
    python_requires='>3.7',
    zip_safe=True,
)



from .event.event import *


def matcher(key: str, *__dict: dict):
    for __d in __dict:
        res = __d.get(key[0].lower() + key[1:], __d.get(key[0].upper() + key[1:]))
        if res:
            return res


def callback_handler(event_body: dict, raw_body=None):
    if raw_body is None:
        raw_body = {}
    if 'EventType' in event_body:
        if event_body['EventType'] == "ChatQuit":
            event = ChatQuit()
            event.time = matcher('Timestamp', event_body)
            event.chatId = matcher('ChatId', event_body)
            event.operatorUnionId = matcher('OperatorUnionId', event_body)
            event.operator = matcher('Operator', event_body)
            event.openConversationId = OpenConversationId(
                matcher('OpenConversationId', event_body))
            event.cropId = matcher('CropId', event_body)
            event.dec_mes = event_body
            event.raw_mes = raw_body
            return event
        elif event_body['EventType'] == 'chat_remove_member':
            event = ChatKick()
            event.time = matcher('Timestamp', event_body)
            event.chatId = matcher('ChatId', event_body)
            event.userIds = matcher('userId', event_body)
            event.operatorUnionId = matcher('OperatorUnionId', event_body)
            event.operator = matcher('Operator', event_body)
            event.openConversationId = OpenConversationId(
                matcher('OpenConversationId', event_body))
            event.cropId = matcher('CropId', event_body)
            event.dec_mes = event_body
            event.raw_mes = raw_body
            return event
        elif event_body['EventType'] == 'chat_update_title':
            event = GroupNameChange()
            event.time = matcher('Timestamp', event_body)
            event.chatId = matcher('ChatId', event_body)
            event.operatorUnionId = matcher('OperatorUnionId', event_body)
            event.operator = matcher('Operator', event_body)
            event.title = matcher('Title', event_body)
            event.openConversationId = OpenConversationId(
                matcher('OpenConversationId', event_body), event.title)
            event.dec_mes = event_body
            event.raw_mes = raw_body
            return event
    event = BasicEvent()
    event.dec_mes = event_body
    event.raw_mes = raw_body
    return event



import hashlib

from ..element import OpenConversationId


class Webhook:
    url: str
    
    expired_time: float
    
    def __init__(self, url: str = None, expired_time: int = None, origin: dict = None):
        if origin is not None:
            url = origin.get('sessionWebhook')
            expired_time = origin.get('sessionWebhookExpiredTime')
        self.url = url
        self.expired_time = expired_time if 2600000000 > expired_time else expired_time / 1000


class Group:
    
    def __init__(self, id: str = None,
                 name: str = None,
                 send_url: str = None,
                 conversationId: str = None,
                 limit_time: int = 0,
                 origin: dict = None):
        if origin is not None:
            id = origin.get('conversationId')
            name = origin.get("conversationTitle")
            send_url = origin.get('sessionWebhook')
            conversationId = origin.get('conversationId')
            limit_time = origin.get('sessionWebhookExpiredTime') or 0
        self.origin_id = id[id.rfind("$"):]
        """原始加密"""
        self.id = (int(hashlib.sha1(self.origin_id.encode('utf-8')).hexdigest(), 16)) % (10 ** 10) + 1000
        """计算所得ID，非真实群号"""
        self.name = name
        """群聊的真实名称"""
        self.webhook = Webhook(send_url, limit_time)
        """群聊的临时Webhook地址，含有URL和过期时间戳"""
        self.conversationId: OpenConversationId = OpenConversationId(conversationId, self.name, self.id)
        """对话ID"""
    
    def __int__(self) -> int:
        return self.id
    
    def __str__(self) -> str:
        return self.name


class Member:
    
    def __init__(self, id: str = None,
                 staffid: str = None,
                 name: str = None,
                 group: Group = None,
                 admin: bool = None,
                 origin: dict = None):
        if origin is not None:
            id = origin.get('senderId')
            name = origin.get("senderNick")
            staffid = origin.get("senderStaffId") or origin.get("StaffId")
            admin = origin.get("isAdmin")
        self.origin_id = id[id.rfind("$"):]
        self.id = (int(hashlib.sha1(self.origin_id.encode('utf-8')).hexdigest(), 16)) % (10 ** 10) + 1000
        self.name = name
        self.staffid = staffid
        self.group = group
        self.admin = admin
    
    def __int__(self) -> int:
        return self.id
    
    def __str__(self) -> str:
        return self.name


class Bot:
    
    def __init__(self, id: str = None, corp_id: str = None, robot_code: str = None, origin: dict = None):
        if origin is not None:
            id = origin.get('chatbotUserId')
            corp_id = origin.get("chatbotCorpId") or origin.get('CropId')
            robot_code = origin.get("robotCode")
        self.origin_id = id[id.rfind("$"):]
        self.id = (int(hashlib.sha1(self.origin_id.encode('utf-8')).hexdigest(), 16)) % (10 ** 10) + 1000
        self.corp_id = corp_id
        self.robot_code = robot_code
    
    def __int__(self) -> int:
        return self.id



from ..message.chain import MessageChain
from ..model import *


class BasicMessage:
    type = "BasicMessage"
    
    message: MessageChain
    
    sender: Member


class GroupMessage(BasicMessage):
    type = "GroupMessage"
    
    message_chain: MessageChain
    
    sender: Member
    
    group: Group
    
    bot: Bot



class ListenerSchema:
    
    def __init__(
            self,
            listening_events: list,
            namespace: str = '',
            inline_dispatchers=None,
            decorators=None,
            priority: int = 16,
            extra_priorities=None
    ):
        if extra_priorities is None:
            extra_priorities = {}
        if decorators is None:
            decorators = []
        if inline_dispatchers is None:
            inline_dispatchers = []
        self.listening_events = listening_events
        self.namespace = namespace
        self.inline_dispatchers = inline_dispatchers
        self.decorators = decorators
        self.priority = priority
        self.extra_priorities = extra_priorities



from typing import List

from ..model import *


class BasicEvent:
    type = "BasicEvent"
    
    def __init__(self, raw_mes: str = "", dec_mes: str = ""):
        self.raw_mes = raw_mes
        self.dec_mes = dec_mes


class CheckUrl(BasicEvent):
    type = 'CheckUrl'


class CheckIn(BasicEvent):
    
    member: Member
    
    bot: Bot


class ChatQuit(BasicEvent):
    
    operator: int
    """对应Operator, 为成员的StaffID"""
    
    openConversationId: OpenConversationId
    """对应OpenConversationId, 为对话ID"""
    
    cropId: str
    """对应CorpId"""
    
    time: int
    """TimeStamp*1000"""
    
    chatId: str
    """对应ChatId"""
    
    operatorUnionId = str
    """对应操作的唯一ID"""


class ChatKick(BasicEvent):
    
    operator: int
    """对应Operator, 为成员的StaffID"""
    
    userIds: List[int]
    """对应userId, 为移除成员的StaffID"""
    
    openConversationId: OpenConversationId
    """对应OpenConversationId, 为对话ID"""
    
    cropId: str
    """对应CorpId"""
    
    time: int
    """TimeStamp*1000"""
    
    chatId: str
    """对应ChatId"""
    
    operatorUnionId = str
    """对应操作的唯一ID"""


class GroupNameChange(BasicEvent):
    
    operator: int
    """对应Operator, 为成员的StaffID"""
    
    operatorUnionId: str
    """对应operatorUnionId，为成员的UnionID"""
    
    openConversationId: OpenConversationId
    """对应OpenConversationId, 为对话ID"""
    
    title: str
    """对应title，为更改后的群标题"""
    
    time: int
    """TimeStamp*1000"""
    
    chatId: str
    """对应ChatId"""
